import numpy as np
import numbers
import os
import matplotlib.pyplot as plt
import json

from ImagingReso import _utilities


class Resonance(object):
    
    database = 'ENDF_VIII'

    E_MIN = 1e-5
    E_MAX = 3e3

    stack = {}  # compound, thickness, atomic_ratio of each layer with isotopes information
    stack_sigma = {}  # all the energy and sigma of the isotopes and compounds
    stack_signal = {}  # transmission and attenuation signal for every isotope and compound
    total_signal = {}  # transmission and attenuation of the entire sample
    
    density_lock = {}  # dictionary that will defined the densities locked
    
    energy_max = np.NaN
    energy_min = np.NaN
    energy_step = np.NaN

    def __init__(self, stack={}, energy_max=1, energy_min=0.001, energy_step=0.001):
        """initialize resonance object
        
        Paramters:
        ==========
        stack: dictionary 
          example: {'layer1': {'elements':['Ag','Si], 
                               'atomic_ratio': [1, 2],
                               'thickness': {'value': 0.025,
                                             'units': 'mm',
                                             },
                               'density': {'units': 'g/cm3',
                                           'value': 0.5,
                                           },
                                }
        energy_max: float (default 300) max energy in eV to use in calculation
        energy_min: float (default 0) min energy in eV to use in calculation
        energy_step: float (default 0.1) energy step to use in extrapolation of sigma data
        """
        self.__element_metadata = {}
        
        if energy_min < self.E_MIN:
            raise ValueError("Energy min (eV) must be >= {}".format(self.E_MIN))
        self.energy_min = energy_min

        if energy_max > self.E_MAX:
            raise ValueError("Energy max (eV) must be <= {}".format(self.E_MAX))
        self.energy_max = energy_max

        self.energy_step = energy_step

        if not stack == {}:
            # checking that every element of each stack is defined
            _utilities.checking_stack(stack=stack)
            new_stack = self.__update_stack_with_isotopes_infos(stack=stack)
            self.stack = new_stack
            
            # if layer density has been defined, lock it
            self.__lock_density_if_defined(stack=self.stack)
            
            # calculate stack_sigma, layer density, atoms_per_cm3 ...
            self.__math_on_stack()
              
    def __str__(self):
        '''what to display if user does
        
        >>> o_reso = Resolution()
        >>> print(o_reso)
        '''
        return json.dumps(self.stack, indent=4)
    
    def __repr__(self):
        '''what to display if user does
        
        >>> o_reso = Resolution()
        >>> o_reso
        '''
        return json.dumps(self.stack, indent=4)    
                
    def add_layer(self, formula='', thickness=np.NaN, density=np.NaN): 
        """provide another way to define the layers (stack)
        
        Parameters:
        ===========
        formula: string
           ex: 'CoAg2'
           ex: 'Al'
        thickness: float (in mm) 
        density: float (g/cm3)
        """
        if formula == '':
            return
        
        _new_stack = _utilities.formula_to_dictionary(formula=formula, 
                                                      thickness=thickness, 
                                                      density=density,
                                                      database=self.database)      
        # check if density has been defined
        self.__lock_density_if_defined(stack=_new_stack)
        
        new_stack = self.__update_stack_with_isotopes_infos(stack=_new_stack)
        self.stack = {**self.stack, **new_stack}
        
        # calculate stack_sigma, layer density, atoms_per_cm3 ...
        self.__math_on_stack()

    def get_isotopic_ratio(self, compound='', element=''):
        """returns the list of isotopes for the element of the compound defined with their stoichiometric values
        
        Parameters:
        ===========
        compound: string (default is empty). If empty, all the stoichiometric will be displayed
        element: string (default is same as compound). 
        
        Raises:
        =======
        ValueError if element is not defined in the stack
        """
        _stack = self.stack
        compound = str(compound)
        
        if compound == '':
            _list_compounds = _stack.keys()
            list_all_dict = {}
            for _compound in _list_compounds:
                _compound = str(_compound)
                _list_element = _stack[_compound]['elements']
                list_all_dict[_compound] = {}
                for _element in _list_element:
                    list_all_dict[_compound][_element] = self.get_isotopic_ratio(
                        compound=_compound,
                        element=_element)
            return list_all_dict
        
        # checking compound is valid
        list_compounds = _stack.keys()
        if not compound in list_compounds:
            list_compounds_joined = ', '.join(list_compounds)
            raise ValueError("Compound '{}' could not be find in {}".format(compound, list_compounds_joined))
        
        # checking element is valid
        if element == '': 
            # we assume that the element and compounds names matched
            element = compound
        list_element = _stack[compound].keys()
        if not element in list_element:
            list_element_joined = ', '.join(list_element)
            raise ValueError("Element '{}' should be any of those elements: {}".format(element, list_element_joined))
        
        list_istopes = _stack[compound][element]['isotopes']['list']
        list_ratio = _stack[compound][element]['isotopes']['isotopic_ratio']
        iso_ratio = zip(list_istopes, list_ratio)
        
        _stoichiometric_ratio = {}
        for _iso, _ratio in iso_ratio:
            _stoichiometric_ratio[_iso] = _ratio
            
        return _stoichiometric_ratio
        
    def set_isotopic_ratio(self, compound='', element='', list_ratio=[]):
        """defines the new set of ratio of the compound/element and trigger the calculation to update the density
        
        Parameters:
        ===========
        compound: string (default is ''). Name of compound
        elememnt: string (defualt is ''). Name of element
        list_ratio: list (default is []). list of new stoichiometric_ratio
        
        Raises:
        =======
        ValueError if compound does not exist
        ValueError if element does not exist
        ValueError if list_ratio does not have the right format
        """
        _stack = self.stack
        
        list_compounds = _stack.keys()
        if not compound in _stack.keys():
            list_compounds_joined = ', '.join(list_compounds)
            raise ValueError("Compound '{}' could not be find in {}".format(compound, list_compounds_joined))

        if element == '': 
            # we assume that the element and compounds names matched
            element = compound
        list_element = _stack[compound].keys()
        if not element in list_element:
            list_element_joined = ', '.join(list_element)
            raise ValueError("Element '{}' should be any of those elements: {}".format(element, list_element_joined))
        
        old_list_ratio = _stack[compound][element]['isotopes']['list']
        if not (len(old_list_ratio) == len(list_ratio)):
            raise ValueError("New list of ratio ({} elements) does not match old list size ({} elements!".format(len(
                list_ratio), len(old_list_ratio)))
        
        self.stack[compound][element]['isotopes']['isotopic_ratio'] = list_ratio
        self.__update_molar_mass(compound=compound, element=element)
        self.__update_density(compound=compound, element=element)

        # update entire stack
        self.__math_on_stack()

    def get_density(self, compound='', element=''):
        """returns the list of isotopes for the element of the compound defined with their density
        
        Parameters:
        ===========
        compound: string (default is empty). If empty, all the stoichiometric will be displayed
        element: string (default is same as compound). 
        
        Raises:
        =======
        ValueError if element is not defined in the stack
        """
        _stack = self.stack
        
        if compound == '':
            _list_compounds = _stack.keys()
            list_all_dict = {}
            for _compound in _list_compounds:
                _list_element = _stack[_compound]['elements']
                list_all_dict[_compound] = {}
                for _element in _list_element:
                    list_all_dict[_compound][_element] = self.get_density(
                        compound=_compound,
                        element=_element)
            return list_all_dict
        
        # checking compound is valid
        list_compounds = _stack.keys()
        if not compound in list_compounds:
            list_compounds_joined = ', '.join(list_compounds)
            raise ValueError("Compound '{}' could not be find in {}".format(compile, list_compounds_joined))
        
        # checking element is valid
        if element == '': 
            # we assume that the element and compounds names matched
            element = compound
        list_element = _stack[compound].keys()
        if not element in list_element:
            list_element_joined = ', '.join(list_element)
            raise ValueError("Element '{}' should be any of those elements: {}".format(element, list_element_joined))
        
        return _stack[compound][element]['density']['value']

    def __set_density(self, compound='', element='', density=np.NaN, debug=False):
        '''defines the new density of the compound/element 
        
        Parameters:
        ===========
        compound: string (default is ''). Name of compound
        element: string (defualt is ''). Name of element
        density: float (default is np.NaN). New density
        
        Raises:
        =======
        ValueError if compound does not exist
        ValueError if element does not exist
        ValueError if density is not a number
        '''
        _stack = self.stack
        
        list_compounds = _stack.keys()
        if not compound in _stack.keys():
            list_compounds_joined = ', '.join(list_compounds)
            raise ValueError("Compound '{}' could not be find in {}".format(compile, list_compounds_joined))

        if element == '': 
            # we assume that the element and compounds names matched
            element = compound
        list_element = _stack[compound].keys()
        if not element in list_element:
            list_element_joined = ', '.join(list_element)
            raise ValueError("Element '{}' should be any of those elements: {}".format(element, list_element_joined))
        
        if not isinstance(density, numbers.Number):
            raise ValueError("Density '{}' must be a number!".format(density))
        
        _density_lock = self.density_lock
        if _density_lock[compound]: 
            raise IOError("You are not authorized to change the density!")

        self.stack[compound]['density']['value'] = density
    
        # populate atoms_per_cm3
        self.__calculate_atoms_per_cm3(used_lock=True)

        # calculate transmission and attenuation
        self.__calculate_transmission_attenuation()
        
    def __math_on_stack(self, used_lock=False):
        """will perform all the various update of the stack, such as populating the stack_sigma, caluclate the density of the
        layers....etc. """

        # populate stack_sigma (Sigma vs Energy for every element)
        self.__get_sigmas()

        # populate compound density (if none provided)
        self.__update_layer_density()

        # populate atoms_per_cm3
        self.__calculate_atoms_per_cm3(used_lock=used_lock)
        
        # calculate transmission and attenuation
        self.__calculate_transmission_attenuation()

    def __lock_density_if_defined(self, stack={}):
        '''lock (True) the density lock if the density has been been defined during initialization
        Store the resulting dictionary into density_lock

        Parameters:
        ===========
        stack: dictionary (optional)
          if not provided, the entire stack will be used
        '''
          
        if self.stack == {}:
            density_lock = {}
        else:
            density_lock = self.density_lock
        
        for _compound in stack.keys():
            _density = stack[_compound]['density']['value']
            if np.isnan(_density):
                density_lock[_compound] = False
            else:
                density_lock[_compound] = True
        self.density_lock = density_lock

    def __calculate_transmission_attenuation(self):
        """  """
        stack = self.stack
        stack_sigma = self.stack_sigma
        stack_signal = {}
        
        total_signal = {}
        total_transmisison = 1.

        # compound level
        for _name_of_compound in stack.keys():
            stack_signal[_name_of_compound] = {}
            transmission_compound = 1.
            energy_compound = []
            
            _list_element = stack[_name_of_compound]['elements']
            _thickness_cm = _utilities.set_distance_units(value=stack[_name_of_compound]['thickness']['value'],
                                                          from_units=stack[_name_of_compound]['thickness']['units'],
                                                          to_units='cm')

            # element level
            for _element in _list_element:
                stack_signal[_name_of_compound][_element] = {}
                _atoms_per_cm3 = stack[_name_of_compound]['atoms_per_cm3'][_element]

                # isotope level
                for _iso in stack[_name_of_compound][_element]['isotopes']['list']:
                    stack_signal[_name_of_compound][_element][_iso] = {}
                    _sigma_iso = stack_sigma[_name_of_compound][_element][_iso]['sigma_b']
                    _transmission_iso = _utilities.calculate_transmission(
                        thickness_cm=_thickness_cm, 
                        atoms_per_cm3=_atoms_per_cm3,
                        sigma_b=_sigma_iso)
                    stack_signal[_name_of_compound][_element][_iso]['transmission'] = _transmission_iso
                    stack_signal[_name_of_compound][_element][_iso]['attenuation'] = 1. - _transmission_iso
                    stack_signal[_name_of_compound][_element][_iso]['energy_eV'] = stack_sigma[_name_of_compound][_element][_iso]['energy_eV']
                
                _sigma_ele = stack_sigma[_name_of_compound][_element]['sigma_b']
                _transmission_ele = _utilities.calculate_transmission(
                    thickness_cm=_thickness_cm, 
                    atoms_per_cm3=_atoms_per_cm3, 
                    sigma_b=_sigma_ele)
                stack_signal[_name_of_compound][_element]['transmission'] = _transmission_ele
                stack_signal[_name_of_compound][_element]['attenuation'] = 1. - _transmission_ele
                stack_signal[_name_of_compound][_element]['energy_eV'] = stack_sigma[_name_of_compound][_element]['energy_eV']

                transmission_compound *= _transmission_ele
                if energy_compound == []:
                    energy_compound = stack_sigma[_name_of_compound][_element]['energy_eV']

            stack_signal[_name_of_compound]['transmission'] = transmission_compound
            stack_signal[_name_of_compound]['attenuation'] = 1. - transmission_compound
            stack_signal[_name_of_compound]['energy_eV'] = energy_compound
            
            total_transmisison *= transmission_compound
        
        total_attenuation = 1. - total_transmisison

        self.stack_signal = stack_signal
        total_signal['transmission'] = total_transmisison
        total_signal['attenuation'] = total_attenuation
        total_signal['energy_eV'] = energy_compound
        self.total_signal = total_signal
        
    def __calculate_atoms_per_cm3(self, used_lock=False):
        """calculate for each element, the atoms per cm3"""
        stack = self.stack
        _density_lock = self.density_lock
        
        for _name_of_compound in stack.keys():
            if used_lock and _density_lock[_name_of_compound]:
                continue
            atoms_per_cm3 = _utilities.get_atoms_per_cm3_of_layer(compound_dict=stack[_name_of_compound])
            stack[_name_of_compound]['atoms_per_cm3'] = atoms_per_cm3
        self.stack = stack

    def __fill_missing_keys(self, stack={}):
        _list_key_to_check = ['density']
        _list_key_value = [{'value': np.NaN,
                            'units': 'g/cm3'}]

        list_compound = stack.keys()
        for _key in list_compound:
            _inside_keys = stack[_key].keys()
            _key_value_to_search = zip(_list_key_to_check, _list_key_value)
            for _key_to_find, _value_to_add in _key_value_to_search:
                if not (_key_to_find in _inside_keys):
                    stack[_key][_key_to_find] = _value_to_add.copy()

        return stack
        
    def __update_layer_density(self, debug=False):
        '''calculate or update the layer density'''
        _stack = self.stack
        
        _density_lock = self.density_lock

        list_compound = _stack.keys()
        for _key in list_compound:
            if _density_lock[_key]:
                continue
            
            #if np.isnan(_stack[_key]['density']['value']):
                
            _list_ratio = _stack[_key]['stoichiometric_ratio']
            _list_density = []
            for _element in _stack[_key]['elements']:
                _list_density.append(_stack[_key][_element]['density']['value'])
                _compound_density = _utilities.get_compound_density(list_density=_list_density, 
                                                                    list_ratio=_list_ratio)   

            _stack[_key]['density']['value'] = _compound_density
        self.stack = _stack
                
    def __update_stack_with_isotopes_infos(self, stack={}):
        """retrieve the isotopes, isotopes file names, mass and atomic_ratio from each element in stack"""
        for _key in stack:
            isotopes_array = []
            _elements = stack[_key]['elements']
            for _element in _elements:
                _dict = _utilities.get_isotope_dicts(element=_element, database=self.database)
                stack[_key][_element] = _dict    
        
        stack = self.__fill_missing_keys(stack=stack)
        return stack

    def __update_density(self, compound='', element=''):
        """Re-calculate the density of the element given due to stoichiometric changes as
        well as the compound density (if density is not locked)
        
        Parameters:
        ===========
        compound: string (default is '') name of compound
        element: string (default is '') name of element
        """
        _density_element = 0
        list_ratio = self.stack[compound][element]['isotopes']['isotopic_ratio']
        list_density = self.stack[compound][element]['isotopes']['density']['value']
        ratio_density = zip(list_ratio, list_density)
        for _ratio, _density in ratio_density:
            _density_element += np.float(_ratio) * np.float(_density)
        self.stack[compound][element]['density']['value'] = _density_element
        
        _density_lock = self.density_lock
        if not _density_lock[compound]:
            self.__update_layer_density()
        
    def __update_molar_mass(self, compound='', element=''):
        """Re-calculate the molar mass of the element given due to stoichiometric changes

        Parameters:
        ==========
        compound: string (default is '') name of compound
        element: string (default is '') name of element
        """
        _molar_mass_element = 0
        list_ratio = self.stack[compound][element]['isotopes']['isotopic_ratio']
        list_mass = self.stack[compound][element]['isotopes']['mass']['value']
        ratio_mass = zip(list_ratio, list_mass)
        for _ratio, _mass in ratio_mass:
            _molar_mass_element += np.float(_ratio) * np.float(_mass)
        self.stack[compound][element]['molar_mass']['value'] = _molar_mass_element
        
    def __get_sigmas(self):
        """will populate the stack_sigma dictionary with the energy and sigma array
        for all the compound/element and isotopes"""
        stack_sigma = {}
        _stack = self.stack

        _file_path = os.path.abspath(os.path.dirname(__file__))
        _database_folder = os.path.join(_file_path, 'reference_data', self.database)

        _list_compounds = _stack.keys()
        for _compound in _list_compounds:
            _list_element = _stack[_compound]['elements']
            stack_sigma[_compound] = {}
            
            for _element in _list_element:
                stack_sigma[_compound][_element] = {}
                _list_isotopes = _stack[_compound][_element]['isotopes']['list']
                _list_file_names = _stack[_compound][_element]['isotopes']['file_names']
                _list_isotopic_ratio = _stack[_compound][_element]['isotopes']['isotopic_ratio']
                _iso_file_ratio = zip(_list_isotopes, _list_file_names, _list_isotopic_ratio)
                
                # _dict_sigma_isotopes_sum = {}
                _sigma_all_isotopes = 0
                _energy_all_isotpes = 0
                
                for _iso, _file, _ratio in _iso_file_ratio:
                    stack_sigma[_compound][_element][_iso] = {}
                    _file = os.path.join(_database_folder, _file)
                    _dict = _utilities.get_sigma(database_file_name=_file, 
                                                 E_min=self.energy_min,
                                                 E_max=self.energy_max,
                                                 E_step=self.energy_step)
                    stack_sigma[_compound][_element][_iso]['energy_eV'] = _dict['energy_eV']
                    stack_sigma[_compound][_element][_iso]['sigma_b'] = _dict['sigma_b'] * _ratio

                    # sigma for all isotopes with their isotopic ratio
                    _sigma_all_isotopes += _dict['sigma_b'] * _ratio
                    _energy_all_isotpes += _dict['energy_eV']

                # energy axis (x-axis) is averaged to take into account differences between x-axis of isotopes
                _mean_energy_all_isotopes = _energy_all_isotpes/len(_list_isotopes)
                stack_sigma[_compound][_element]['energy_eV'] = _mean_energy_all_isotopes
                stack_sigma[_compound][_element]['sigma_b'] = _sigma_all_isotopes
                    
        self.stack_sigma = stack_sigma

    def plot(self, transmission=False, x_axis='energy', mixed=True, all_layers=False, all_elements=False,
             all_isotopes=False, items_to_plot=[], time_unit='us', offset_us=2.99, time_resolution_us=0.16, source_to_detector_m=16.125):
        # offset delay values is normal 2.99 us with NONE actual MCP delay settings
        """display the transmission or attenuation of compound, element and/or isotopes specified

        Parameters:
        ===========
        transmission: boolean. True -> display transmission signal
                               False -> display the attenuation signal
        x_axis: string. Must be either 'energy' or 'lambda'
        mixed: boolean. True -> display the total of each layer
                        False -> not displayed
        all_layers: boolean. True -> display all layers
                             False -> not displayed
        all_elements: boolean. True -> display all elements signal
                               False -> not displayed
        all_isotopes: boolean. True -> display all isotopes signal
                               False -> not displayed
        items_to_plot: array that descibes what to plot
            ex:
                [['CoAg','Ag','107-Ag'], ['CoAg']]
            if the dictionary is empty, everything is plotted
        """
        plt.figure(figsize=[8, 8])

        _stack_signal = self.stack_signal
        _stack = self.stack
        _x_axis = self.total_signal['energy_eV']

        '''X-axis'''
        # determine values and labels for x-axis with options from
        # 'energy(eV)' & 'lambda(A)' & 'time(us)' & 'image number(#)'
        if x_axis == 'energy':
            x_axis_label = 'Energy (eV)'
        if x_axis == 'lambda':
            x_axis_label = u"Wavelength (\u212B)"
            _x_axis = _utilities.ev_to_angstroms(array=_x_axis)
            # plt.xlim(0, 1)
        if x_axis == 'time':
            if time_unit == 's':
                x_axis_label = 'Time (s)'
                _x_axis = _utilities.ev_to_s(array=_x_axis,
                                             source_to_detector_m=source_to_detector_m,
                                             offset_us=offset_us)
            if time_unit == 'us':
                x_axis_label = 'Time (us)'
                _x_axis = 1e6 * _utilities.ev_to_s(array=_x_axis,
                                                   source_to_detector_m=source_to_detector_m,
                                                   offset_us=offset_us)
            if time_unit == 'ns':
                x_axis_label = 'Time (ns)'
                _x_axis = 1e9 * _utilities.ev_to_s(array=_x_axis,
                                                   source_to_detector_m=source_to_detector_m,
                                                   offset_us=offset_us)
        if x_axis == 'number':
            x_axis_label = 'Image number (#)'
            _x_axis = _utilities.ev_to_image_number(array=_x_axis,
                                                    source_to_detector_m=source_to_detector_m,
                                                    offset_us=offset_us,
                                                    time_resolution_us=time_resolution_us)

        '''Y-axis'''
        # determine to plot transmission or attenuation
        # determine to put transmission or attenuation words for y-axis
        if transmission:
            y_axis_label = 'Neutron Transmission'
            y_axis_tag = 'transmission'
        else:
            y_axis_label = 'Neutron Attenuation'
            y_axis_tag = 'attenuation'

        if mixed:
            _y_axis = self.total_signal[y_axis_tag]
            plt.plot(_x_axis, _y_axis, label="Total")

        if all_layers:
            for _compound in _stack.keys():
                _y_axis = _stack_signal[_compound][y_axis_tag]
                plt.plot(_x_axis, _y_axis, label=_compound)

        if all_elements:
            for _compound in _stack.keys():
                for _element in _stack[_compound]['elements']:
                    _y_axis = _stack_signal[_compound][_element][y_axis_tag]
                    plt.plot(_x_axis, _y_axis, label="{}/{}".format(_compound, _element))

        if all_isotopes:
            for _compound in _stack.keys():
                for _element in _stack[_compound]['elements']:
                    for _isotope in _stack[_compound][_element]['isotopes']['list']:
                        _y_axis = _stack_signal[_compound][_element][_isotope][y_axis_tag]
                        plt.plot(_x_axis, _y_axis, label="{}/{}/{}".format(_compound, _element, _isotope))

        '''Y-axis for specified items_to_plot'''
        for _path_to_plot in items_to_plot:
            _path_to_plot = list(_path_to_plot)
            _live_path = _stack_signal
            _label = "/".join(_path_to_plot)
            while _path_to_plot:
                _item = _path_to_plot.pop(0)
                _live_path = _live_path[_item]
            _y_axis = _live_path[y_axis_tag]
            plt.plot(_x_axis, _y_axis, label=_label)

        plt.ylim(-0.01, 1.01)
        plt.xlabel(x_axis_label)
        plt.ylabel(y_axis_label)
        plt.legend(loc='best')
        plt.show()



import glob
import os
import numbers
import re
import numpy as np
import periodictable as pt
import pandas as pd
from scipy.interpolate import interp1d
from scipy.constants import Avogadro


def is_element_in_database(element='', database='ENDF_VIII'):
    '''will try to find the element in the folder (database) specified
    
    Parameters:
    ==========
    element: string. Name of the element. Not case sensitive
    database: string (default is 'ENDF_VIII'). Name of folder that has the list of elements
    
    Returns:
    =======
    bool: True if element was found in the database
          False if element could not be found
    '''
    if element == '':
        return False
    
    list_entry_from_database = get_list_element_from_database(database=database)
    if element.lower() in list_entry_from_database:
        return True
    return False    
    
def get_list_element_from_database(database=''):
    '''return a string array of all the element from the database
    
    Parameters:
    ==========
    database: string. Name of database
    
    Raises:
    ======
    ValueError if database can not be found
    
    '''
    _file_path = os.path.abspath(os.path.dirname(__file__))
    _database_folder = os.path.join(_file_path, 'reference_data', database)

    if not os.path.exists(_database_folder):
        raise ValueError("Database {} does not exist!".format(database))
    
    _list_files = glob.glob(_database_folder + '/*.csv')
    _list_short_files = [os.path.basename(_file) for _file in _list_files]
    _list_element = set([_name.split('-')[0].lower() for _name in _list_short_files])
    return _list_element
    
def checking_stack(stack={}, database='ENDF_VIII'):
    '''This method makes sure that all the elements from the various stacks are 
    in the database and that the thickness has the correct format (float)
    
    Parameters:
    ==========
    stack: dictionary that defines the various stacks
    database: string (default is 'ENDF_VIII') name of database
    
    Raises:
    ======
    ValueError if one of the element in one of the stack can not be found 
    ValueError if thickness is not a float
    ValueError if elements and stoichiometric ratio arrays do not have the same size
    
    Return:
    ======
    True: for testing purpose only
    '''
    for _keys in stack:
        _elements = stack[_keys]['elements']
        for _element in _elements:
            if not is_element_in_database(element=_element):
                raise ValueError("Element {} can not be found in the database".format(_element))

        _thickness = stack[_keys]['thickness']['value']
        if not isinstance(_thickness, numbers.Number):
            raise ValueError("Thickness {} should be a number!".format(_thickness))
    
        _stoichiometric_ratio = stack[_keys]['stoichiometric_ratio']
        if len(_stoichiometric_ratio) != len(_elements):
            raise ValueError("stoichiometric Ratio and Elements should have the same size!")
    
    return True    

def formula_to_dictionary(formula='', thickness=np.NaN, density=np.NaN, database='ENDF_VIII'):
    '''create dictionary based on formula given
    
    Parameters:
    ===========
    formula: string
       ex: 'AgCo2'
       ex: 'Ag'
    thickness: float (in mm) default is np.NaN
    density: float (in g/cm3) default is np.NaN
    database: string (default is ENDV_VIII). Database where to look for elements
    
    Raises:
    =======
    ValueError if one of the element is missing from the database
    
    Return:
    =======
    the dictionary of the elements passed
      ex: {'AgCo2': {'elements': ['Ag','Co'],
                     'stoichiometric_ratio': [1,2],
                     'thickness': {'value': thickness,
                                   'units': 'mm',
                                   },
                     'density': {'value': density,
                                 'units': 'g/cm3',
                                 },
                    }
    '''
    _formula_parsed = re.findall(r'([A-Z][a-z]*)(\d*)', formula)

    _dictionary = {}
    _elements_array = []
    _atomic_ratio_array = []
    for _element in _formula_parsed:
        [_single_element, _atomic_ratio] = list(_element)
        if not is_element_in_database(element=_single_element, database=database):
            raise ValueError("element {} not found in database!".format(_single_element))

        if _atomic_ratio == '':
            _atomic_ratio = 1

        _atomic_ratio_array.append(int(_atomic_ratio))
        _elements_array.append(_single_element)

    return {formula: {'elements': _elements_array,
                      'stoichiometric_ratio': _atomic_ratio_array,
                      'thickness': {'value': thickness,
                                    'units': 'mm'},
                      'density': {'value': density,
                                  'units': 'g/cm3',
                                  }
                      },
            }

def get_isotope_dicts(element='', database='ENDF_VIII'):
    '''return a dictionary with list of isotopes found in database and name of database files
    
    Parameters:
    ===========
    element: string. Name of the element
      ex: 'Ag'
    database: string (default is ENDF_VIII)
    
    Returns:
    ========
    dictionary with isotopes and files 
      ex: {'Ag': {'isotopes': ['107-Ag','109-Ag'],
                  'file_names': ['Ag-107.csv','Ag-109.csv']}}
    
    '''
    _file_path = os.path.abspath(os.path.dirname(__file__))
    _database_folder = os.path.join(_file_path, 'reference_data', database)
    _element_search_path = os.path.join(_database_folder, element + '-*.csv')
    list_files = glob.glob(_element_search_path)
    isotope_dict = {'isotopes': {'list': [], 
                                 'file_names': [],
                                 'density': {'value': np.NaN,
                                             'units': 'g/cm3'},
                                 'mass': {'value': [],
                                          'units': 'g/mol',
                                 },
                                 'isotopic_ratio': [],},
                    'density': {'value': np.NaN,
                                'units': 'g/cm3'},
                    'molar_mass': {'value': np.NaN,
                                   'units': 'g/mol'},
                    } 

    isotope_dict_mirror = {}
    _isotopes_list = []
    _isotopes_list_files = []
    _isotopes_mass = []
    _isotopes_density = []
    _isotopes_atomic_ratio = []
    _density = np.NaN
    _molar_mass = np.NaN
    
    for file in list_files:

        # Obtain element, z number from the basename
        _basename = os.path.basename(file)
        [filename, file_extension] = os.path.splitext(_basename)
        [_name, _number] = filename.split('-')
        _symbol = _number + '-' + _name
        isotope = str(_symbol)

        _isotopes_list.append(isotope)
        _isotopes_list_files.append(_basename)
        _isotopes_mass.append(get_mass(isotope))
        _isotopes_atomic_ratio.append(get_abundance(isotope))
        _isotopes_density.append(get_density(isotope))
        _density = get_density(element)
        _molar_mass = get_mass(element)
                                        
    isotope_dict['isotopes']['list'] = _isotopes_list
    isotope_dict['isotopes']['file_names'] = _isotopes_list_files              
    isotope_dict['isotopes']['mass']['value'] = _isotopes_mass
    isotope_dict['isotopes']['isotopic_ratio'] = _isotopes_atomic_ratio
    isotope_dict['isotopes']['density']['value'] = _isotopes_density
    isotope_dict['density']['value'] = _density
    isotope_dict['molar_mass']['value'] = _molar_mass
                    
    return isotope_dict   

def get_abundance(element):
    '''return the abundance [0.,1.] of the defined element
    
    Parameters:
    ===========
    element: (sting)
    
    Returns:
    ========
    the abundance of the elemenet (value between 0 and 1)
    '''
    return pt.elements.isotope(element).abundance / 100.

def get_mass(element):
    '''return the molar mass (SI units) of an given isotope, or element
    
    Parameters:
    ===========
    element: string. Element or isotopes to get the mass from
    
    Returns:
    ========
    the molar mass of the element/isotope in SI units
    '''
    return pt.elements.isotope(element).mass

def get_density(element):
    '''return the density (g.cm-3) of the element
    
    Paramters:
    ==========
    element: string. Name of element
    
    Returns:
    ========
    the density of the element in g.cm-3 units
    '''
    return pt.elements.isotope(element).density

def get_compound_density(list_density=[], list_ratio=[]):
    ''''''
    _ratio_density = zip(list_ratio, list_density)
    _density_compound = 0
    
    _sum_ratio = np.array(list_ratio).sum()
    
    for _ratio, _density in _ratio_density:
        _density_compound += (_ratio * _density) / _sum_ratio
    return _density_compound

def get_database_data(file_name=''):
    '''return the energy (eV) and Sigma (barn) from the file_name
    
    Parameters:
    ===========
    file_name: string ('' by default) name of csv file
    
    Returns:
    ========
    pandas dataframe
    
    Raises:
    =======
    IOError if file does not exist
    '''
    if not os.path.exists(file_name):
        raise IOError("File {} does not exist!".format(file_name))
    df = pd.read_csv(file_name, header=1)
    return df

def get_interpolated_data(df=[], E_min=np.NaN, E_max=np.NaN, E_step=np.NaN):
    '''return the interpolated x and y axis for the given x range [E_min, E_max] with step defined
    
    Parameters:
    ===========
    df: data frame 
    E_min: left range of new interpolated data
    E_max: right range of new interpolated data
    E_step: step of energy to use in interpolated data
    
    Returns:
    ========
    x_axis and y_axis of interpolated data over specified range
    '''
    nbr_point = (E_max - E_min) / E_step
    
    # remove data outside specified range [x_min, x_max]
    #df = df.drop(df[df.E_eV < E_min].index)
    #df = df.drop(df[df.E_eV > E_max].index)

    # reset index
    df = df.reset_index(drop=True)
    
    # energy x_axis
    #x_axis = np.linspace(df['E_eV'].min(), df['E_eV'].max(), nbr_point)
    x_axis = np.linspace(E_min, E_max, nbr_point)
    y_axis_function = interp1d(x=df['E_eV'], y=df['Sig_b'], kind='linear')
    
    y_axis = y_axis_function(x_axis)   
    
    return {'x_axis': x_axis, 'y_axis': y_axis}

def get_sigma(database_file_name='', E_min=np.NaN, E_max=np.NaN, E_step=np.NaN):
    '''retrieve the Energy and sigma axis for the given isotope
    
    Paramters:
    ==========
    database_file_name: string
    E_min: left range of new interpolated data
    E_max: right range of new interpolated data
    E_step: step of energy to use in interpolated data
    
    Returns:
    ========
    {'energy': np.array(), 'sigma': np.array}
    '''
    _df = get_database_data(file_name=database_file_name)
    _dict = get_interpolated_data(df=_df, E_min=E_min, E_max=E_max, 
                                 E_step=E_step)
    return {'energy_eV': _dict['x_axis'],
            'sigma_b': _dict['y_axis']}

def get_atoms_per_cm3_of_layer(compound_dict={}):
    '''calculate the atoms per cm3 of the given compound (layer)
    
    Paramters:
    ==========
    compound_dict: {} 
    
    Returns:
    ========
    dictionary
    '''
    atoms_per_cm3 = {}
    
    _list_of_elements = compound_dict['elements']
    _stoichiometric_list = compound_dict['stoichiometric_ratio']

    _element_stoichio = zip(_list_of_elements, _stoichiometric_list)
    _molar_mass_sum = 0
    for _element, _stoichio in _element_stoichio:
        _molar_mass_sum += _stoichio * compound_dict[_element]['molar_mass']['value']

    _element_stoichio = zip(_list_of_elements, _stoichiometric_list)
    for _element, _stoichio in _element_stoichio:
        _step1 = (compound_dict['density']['value'] * _stoichio) / _molar_mass_sum
        atoms_per_cm3[_element] = Avogadro * _step1
    
    return atoms_per_cm3

def calculate_transmission(thickness_cm=np.NaN, atoms_per_cm3=np.NaN, sigma_b=[]):
    '''calculate the transmission signal using the formula
    
    transmission = exp( - thickness_cm * atoms_per_cm3 * 1e-24 * sigma_b)
    
    Parameters:
    ===========
    thickness: float (in cm)
    atoms_per_cm3: float (number of atoms per cm3 of element/isotope)
    sigma_b: np.array of sigma retrieved from database
    
    Returns:
    ========
    transmission array
    '''
    transmission = np.exp( -thickness_cm * 1e-24 * sigma_b * atoms_per_cm3)
    return np.array(transmission)

def set_distance_units(value=np.NaN, from_units='mm', to_units='cm'):
    '''convert distance into new units
    
    Parameters:
    ===========
    value: float. value to convert
    from_units: string. Must be 'mm', 'cm' or 'm'
    to_units: string. must be 'mm','cm' or 'm'
    
    Returns:
    ========
    converted value
    
    Raises:
    =======
    ValueError if from_units is not a valid unit (see above)
    ValueError if to_units is not a valud unit
    '''
    if from_units == to_units:
        return value

    coeff = 1
    if from_units == 'cm':
        if to_units == 'mm':
            coeff = 10
        elif to_units == 'm':
            coeff = 0.01
        else:
            raise ValueError("to_units not supported ['cm','m','mm']!")
    elif from_units == 'mm':
        if to_units == 'cm':
            coeff = 0.1
        elif to_units == 'm':
            coeff = 0.001
        else:
            raise ValueError("to_units not supported ['cm','m','mm']!")
    elif from_units == 'm':
        if to_units == 'mm':
            coeff = 1000
        elif to_units == 'cm':
            coeff = 100
        else:
            raise ValueError("to_units not supported ['cm','m','mm']!")
    else:
        raise ValueError("to_units not supported ['cm','m','mm']!")

    return coeff * value

def ev_to_angstroms(array=[]):
    """convert into lambda from the energy array

    Parameters:
    ===========
    array: array to convert (in eV)

    Returns:
    ========
    numpy array of lambda in Angstroms
    """
    return np.sqrt(81.787 / (array * 1000.))  # 1000 is used to convert eV to meV

def angstroms_to_ev(array=[]):
    """convert lambda array in angstroms to energy in eV
    
    Parameters:
    ===========
    array: numpy array in Angstroms
    
    Returns:
    ========
    numpy array of energy in eV
    """
    return 81.787 / (1000. * array ** 2)  # 1000 is used to convert meV to eV

def ev_to_s(array=[], offset_us=np.NaN, source_to_detector_m=np.NaN):
    # delay values is normal 2.99 us with NONE actual MCP delay settings
    """convert energy (eV) to time (us)

    Parameters:
    ===========
    array: array (in eV)
    offset_us: float. Delay of detector in us
    source_to_detector_m: float. Distance source to detector in m

    Returns:
    ========
    time: array in s 
    """
    # 1000 is used to convert eV to meV
    time_s = np.sqrt(81.787 / (array * 1000.)) * source_to_detector_m / 3956.
    time_record_s = time_s - offset_us * 1e-6
    return time_record_s

def s_to_ev(array=[], offset_us=np.NaN, source_to_detector_m=np.NaN):
    """convert time (s) to energy (eV)
    Parameters:
    ===========
    numpy array of time in s
    offset_us: float. Delay of detector in us
    source_to_detector_m: float. Distance source to detector in m

    Returns:
    ========
    numpy array of energy in eV
    """
    lambda_a = 3956. * (array + offset_us*1e-6) / source_to_detector_m
    return (81.787 / pow(lambda_a, 2))/1000.  # 1000 is used to convert meV to eV

def angstroms_to_s(array=[], offset_us=np.NaN, source_to_detector_m=np.NaN):
    '''convert array in angstroms into s

    Parameters:
    ===========
    numpy array of lambda in angstroms
    offset_us: float. Delay of detector in us
    source_to_detector_m: float. Distance source to detector in m

    Returns:
    ========
    numpy array of time in s
    '''
    return (source_to_detector_m * array / 3956.) - offset_us * 1e-6

def s_to_angstroms(array=[], offset_us=np.NaN, source_to_detector_m=np.NaN):
    '''convert s to angstroms arrays

    Parameters:
    ===========
    array: array in s
    offset_us: float. Delay of detector in mocros
    source_to_detector_m: float. Distance source to detector in m

    Returns:
    ========
    array in angstroms
    '''
    return 3956. * (array + offset_us*1e-6) / source_to_detector_m

def ev_to_image_number(array=[], offset_us=np.NaN, time_resolution_us=np.NaN, source_to_detector_m=np.NaN, t_start_us = 1):
    # delay values is normal 2.99 us with NONE actual MCP delay settings
    """convert energy (eV) to image numbers (#)

    Parameters:
    ===========
    numpy array of energy in eV
    offset_us: float. Delay of detector in us
    source_to_detector_m: float. Distance source to detector in m

    Returns:
    ========
    image numbers: array of image number
    """
    time_tot_us = np.sqrt(81.787 / (array * 1000)) * source_to_detector_m * 100 / 0.3956
    time_record_us = (time_tot_us + offset_us)
    image_number = (time_record_us - t_start_us) / time_resolution_us
    return image_number


def convert_x_axis(array=[], from_units='ev', to_units='Angstroms',
                   offset_us=np.NaN,
                   source_to_detector_m=np.NaN):
    '''allow to convert the x-axis into eV, Angstroms units, or s
    Parameters:
    ===========
    array: array to convert
    from_units: string (default is eV)
    to_units: string (default is Angstroms)
    offset_us: float. Delay in micros of the detector
    source_to_detector_m: float. distance source detector in m
    Returns:
    ========
    converted array
    '''
    units_allowed = ['ev', 'angstroms', 's']

    if array == []:
        return np.ndarray([])

    _array = array.copy()

    from_units = from_units.lower()
    if not from_units in units_allowed:
        raise ValueError("Units {} not supported!".format(from_units))

    to_units = to_units.lower()
    if not to_units in units_allowed:
        raise ValueError("Units {} not supported!".format(to_units))

    if from_units == to_units:
        return array

    if from_units == 'ev':
        if to_units == 'angstroms':
            return ev_to_angstroms(array=array)

        if to_units == 's':
            if np.isnan(source_to_detector_m):
                raise ValueError("Please provide a distance source-detector in m!")

            return ev_to_s(array=array,
                           offset_us=offset_us,
                           source_to_detector_m=source_to_detector_m)

    if from_units == 'angstroms':
        if to_units == 'ev':
            return angstroms_to_ev(array=array)

        if to_units == 's':
            if np.isnan(source_to_detector_m):
                raise ValueError("Please provide a distance source-detector in m!")

            return angstroms_to_s(array=array,
                                  offset_us=offset_us,
                                  source_to_detector_m=source_to_detector_m)

    if from_units == 's':
        if to_units == 'ev':
            return s_to_ev(array=array,
                           offset_us=offset_us,
                           source_to_detector_m=source_to_detector_m)

        if to_units == 'angstroms':
            return s_to_angstroms(array=array,
                                  offset_us=offset_us,
                                  source_to_detector_m=source_to_detector_m)

    return np.ndarray([])


#!/usr/bin/env python
from setuptools import setup, find_packages
import codecs


def read_file(filename):
    """
    Read a utf8 encoded text file and return its contents.
    """
    with codecs.open(filename, 'r', 'utf8') as f:
        return f.read()

setup(
    name = "ImagingReso",
    version = "1.5.2",
    author = "Yuxuan (Shawn) Zhang, Jean Bilheux",
    author_email = "zhangy6@ornl.gov, bilheuxjm@ornl.gov",
    packages = find_packages(exclude=['tests', 'notebooks']),
    package_data = {'ImagingReso' : ['reference_data/ENDF_VII/*',
                                     'reference_data/ENDF_VIII/*']},
    include_package_data = True,
    test_suite = 'tests',
    install_requires = [
        'numpy',
        'pandas',
        'periodictable',
        'pprint',
        'scipy',
    ],
    dependency_links = [
    ],
    description = "resonance imaging neutron data",
    long_description = read_file('README.md'),
    license = 'BSD',
    keywords = ['neutron','resonance','imaging'],
    url = "https://github.com/ornlneutronimaging/ImagingReso",
    classifiers = ['Development Status :: 3 - Alpha',
                   'License :: OSI Approved :: BSD License',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Intended Audience :: Developers',
                   'Programming Language :: Python', 
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Natural Language :: English'],
)


# End of file
